#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <netinet/in.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <net/if.h>
#include <stdio.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include "net_linux.h"
#include "global.h"
#include "md5.h"

#include <sstream>
#include "QString"

#define logictostr(x) (x) ? "yes" : "no"

#ifndef SOL_PACKET
#warning SOL_PACKET undeclared !
#define SOL_PACKET 263
#endif
#if 0
/* Chez Intellon ils ne sont pas foutus de mettre dans la trame
 * Channel capabilities si les donnees concernent le TX ou le RX.
 * Il faut donc se rappeller de la derniere commande envoyee, et en plus de
 * la derniere commande envoyee par les autres...
 */
unsigned char capabilities_flag = 0;

struct sockaddr_ll addr_broadcast = { 0,0,0,0,0,0,{0,} };
unsigned char outframe[200];
int netfd=0;
unsigned char *framebuf=NULL;
unsigned int buflen;
int initialized = 0;
struct ifreq ifr;

unsigned short ex_word(unsigned char *ptr) {
  return ntohs(*((unsigned short*)ptr));
}

unsigned long ex_long(unsigned char *ptr) {
  return ntohl(*((unsigned long*)ptr));
}

char *format_mac_addr(unsigned char *addr, char *macbuf) {

  sqDebug(macbuf, "%02x:%02x:%02x:%02x:%02x:%02x",
      addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);

  return macbuf;
}
void dump_network_statistics(unsigned char *macmgmt)
{
  int da;
  u_char *stat;
  char macbuf[20];

  for (da = 0; da < 15; da++) {

    stat = (macmgmt+9+da*12);

    /* Check to see if that node entry is valid -
       stupid Intellon chip is supposed to return 00:00:00:00:00:00 for
       nonexistant nodes (as per the specs), but instead it returns
01:00:00:00:00:00 so we just skip checking the first byte,
since heaven knows what else it may return instead of 01 in
other places/revisions.
*/

    if (!((stat[1] == 0) && (stat[2] == 0) &&
	  (stat[3] == 0) && (stat[4] == 0) && (stat[5] == 0))) {

      qDebug("\n  Statistics for Network DA #%d:\n"
	  "  MAC address:         %s\n"
	  "  Bytes in 40 symbols: %u\n"
	  "  FAILS received:      %u\n"
	  "  Frame Drops:         %u\n",

	  da+1, format_mac_addr(stat, macbuf), ex_word(&stat[6]),
	  ex_word(&stat[8]), ex_word(&stat[10])); 
    }
  }
}

void dump_tx_characteristics(u_char *macmgmt) {

  char *retrtab[] = {"Transmit without retries",
    "Transmit with one retry only",
    "Transmit with normal retries (HomePlug)", "Reserved"};

  qDebug("  Local consumption only:        %s\n"
      "  Encryption flag:               %s\n"
      "  Transmit priority:             %u\n"
      "  Response expected:             %s\n"
      "  Transmit contention free:      %s\n"
      "  Retry control:                 %s\n"
      "  No default encryption receive: %s\n"
      "  No unencrypted receive:        %s\n"
      "  Transmit EKS:                  %u\n",

      logictostr(macmgmt[2] & 0x80), logictostr(macmgmt[2] & 0x40),
      (macmgmt[2] >> 4) & 0x03, logictostr(macmgmt[2] & 0x08), 
      logictostr(macmgmt[2] & 0x04), retrtab[(macmgmt[3] >> 6) & 0x03],
      logictostr(macmgmt[3] & 0x08), logictostr(macmgmt[3] & 0x04),
      macmgmt[4]);
}

void dump_set_key(u_char *macmgmt) {

  char	asckey[17];
  char	*hextab = "0123456789abcdef";
  int	i;

  /* Convert the key to ASCII hex */
  for (i = 0; i < 8; i++) {
    asckey[i<<1] = hextab[(macmgmt[i+3] >> 4) & 0x0F];
    asckey[(i<<1)+1] = hextab[macmgmt[i+3] & 0x0F];
  }
  asckey[16] = 0;
  qDebug("\tEncryption key select: 0x%02x\n\tNetwork encryption key: %s\n", macmgmt[2], asckey); 
}

int plcnet_add_if(const QString &if_name)
{
  u_int i;
  struct packet_mreq mreq;

  qDebug("plcnet_add_if : %s\n", if_name.toAscii().constData());

  netfd = socket(PF_PACKET, SOCK_DGRAM, htons(ETHERTYPE_INTELLON));
  if (netfd == -1) {
    perror("socket");
    return -1;
  }

  strncpy(ifr.ifr_name, if_name.toAscii().constData(), sizeof(ifr.ifr_name));
  if (ioctl(netfd, SIOCGIFMTU, &ifr) == -1) {
    perror("ioctl(SIOCGIFMTU)");
    return -1;
  }

  qDebug("Connected on %s\n", ifr.ifr_name);
  buflen = ifr.ifr_mtu + ETHER_HDR_LEN;
  /* Allocate buffer */
  if (!(framebuf = (u_char*)malloc((size_t)buflen))) {
    fqDebug(stderr, "Cannot malloc() packet buffer!\n");
    return -1;
  }

  /* Bind to interface */
  if (ioctl(netfd, SIOCGIFINDEX, &ifr) == -1) {
    perror("ioctl(SIOCGIFINDEX)");
    return -1;
  }
  addr_broadcast.sll_family = AF_PACKET;
  addr_broadcast.sll_protocol = htons(ETHERTYPE_INTELLON);
  addr_broadcast.sll_ifindex = ifr.ifr_ifindex;
  if (bind(netfd, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast)) == -1) {
    perror("bind");
    return -1;
  }

  qDebug("Debug : on passe en PROMISC !\n");
  memset(&mreq, 0, sizeof(mreq));
  mreq.mr_ifindex = ifr.ifr_ifindex;
  mreq.mr_type    = PACKET_MR_PROMISC;

  if (setsockopt(netfd, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) == -1)
  {
    perror("setsockopt(PACKET_ADD_MEMBERSHIP, PROMISC)");
    return -1;
  }

  addr_broadcast.sll_family = AF_PACKET;
  addr_broadcast.sll_protocol = htons(ETHERTYPE_INTELLON);
  addr_broadcast.sll_ifindex = ifr.ifr_ifindex;
  addr_broadcast.sll_halen = 6;
  for (i = 0; i < 6; i++)
    addr_broadcast.sll_addr[i] = 0xFF;
  initialized = 1;
  return netfd;
}

int plc_req_param_stat()
{
  qDebug("send : request parameters & statistics\n");
  outframe[0] = 0x01; /* one MAC management entry */
  outframe[1] = 0x07;	/* request parameters and statistics */
  outframe[2] = 0x0;  /* 0 bytes follow */
  sendto(netfd, outframe, 3, 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
  return 0;
}

int plc_set_network_key(const QString& mac, const QString& key)
{
  // Use PBKDF1 to convert plaintext password
 struct sockaddr_ll addr= { 0,0,0,0,0,0,{0,} };
  int i;
  MD5 myMD5;
  md5_byte_t digest[16];

  /* generate initial digest */
  myMD5.append((const md5_byte_t *)key.toAscii().constData(), key.length());
  myMD5.append((const md5_byte_t *)"\x08\x85\x6d\xaf\x7c\xf5\x81\x85", 8);
  myMD5.finish();
  memcpy(digest, myMD5.getDigest(), 16);

  /* loop 999 times as required by HomePlug */
  for (i = 0; i < 999; i++) {
    myMD5.reset();
    myMD5.append(digest, 16);
    myMD5.finish();
    memcpy(digest, myMD5.getDigest(), 16);
  }

  /* extract the first 8 bytes; calculate parity bit (LSB = odd parity) */
  for (i = 0; i < 8; i++)
  {
    digest[i] = deskeyparity(digest[i]);
  }

  /* MAC Homeplug MME format (p34)
   * MMECTRL   = 01
   * MEHDR = (MEV<<5) + MTYPE  = 0x04
   * MMELength = 09
   * EKS    = 01 (encryption key select)
   * NEK (8 bytes)
   */
  addr.sll_family  = AF_PACKET;
  addr.sll_protocol= htons(ETHERTYPE_INTELLON);
  addr.sll_ifindex = ifr.ifr_ifindex;

  std::wistringstream Stream(mac.toStdWString());

  for (i=0;i < 6; i++)
  {
    unsigned Byte;
    wchar_t Sep;
    if (i > 0)
    {
      Stream >> Sep;
    }
    Stream >> std::hex >> Byte;
    addr.sll_addr[i] = Byte;
  }

  qDebug("plc_send_req_version to ");
  for (i=0; i < 6; ++i)
  {
    qDebug("<%02x>",addr.sll_addr[i]);
  }
  qDebug("\n");
  qDebug("send : set network key to : %s\n", key.toAscii().constData());
  outframe[0] = 0x01; /* one MAC management entry */
  outframe[1] = 0x04;	/* set network key */
  outframe[2] = 0x09;	/* 9 bytes follow */
  outframe[3] = 0x01;	/* encryption key select -> 1 */
  qDebug("NEK = ");
  for (i = 0; i < 8; i++)
  {
    outframe[i+4] = digest[i];
    qDebug("0x%X ", digest[i]);
  }
  qDebug("\n");
  sendto(netfd, outframe, 12, 0, (struct sockaddr *)&addr, sizeof(addr));
  return 0;
}

int plc_req_intellon_specific()
{
  int i;

  qDebug("send : request Intellon-specific network statistics\n");
  outframe[0] = 0x01; /* one MAC management entry */
  outframe[1] = 0x1a;	/* request network statistics */
  outframe[2] = 0xbb;	/* 187 bytes follow */
  outframe[3] = 0x80;	/* read the stats, don't clear them */
  for (i = 0; i < 186; i++)
  {
    outframe[i+4] = 0x00;
  }
  sendto(netfd, outframe, 190, 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
  return 0;
}

int plc_req_perf_stat()
{
  int i;
  qDebug("send : request perf statistics\n");
  outframe[0] = 0x01; /* one MAC management entry */
  outframe[1] = 0x1c;	/* request network statistics */
  outframe[2] = 123;	/* 123 bytes follow */
  outframe[3] = 0x80;	/* read the stats, don't clear them */
  for (i = 0; i < 122; i++)
  {
    outframe[i+4] = 0x00;
  }
  return sendto(netfd, outframe, 190, 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
}

int plc_deinit()
{
  qDebug("plc_deinit\n");
  initialized = 0;
  if (framebuf)
  {
    free(framebuf);
  }
  if (netfd)
  {
    close(netfd);
  }
  return 0;
}

// send mac management
int plc_send_mm()
{
  /*
     unsigned char mme[] = {                                                                0x01, 0x02,
     0x05, 0x00, 0x04, 0x87, 0x10, 0x00, 0x00, 0x00,   0x00, 0x00, 0x00, 0x00, 0xe8, 0x7d, 0x41, 0x00,
     0x01, 0x00, 0x00, 0x00, 0xe0, 0x9F, 0x40, 0x00,   0x40, 0x60, 0x41, 0x00, 0x10, 0x00, 0x00, 0x00,
     0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00,   0xfc, 0x9f, 0x40, 0x00};
     */

  unsigned char mme[] = { 0x01, 0x02, 0x05, 0x00, 0x04, 0x87, 0x10};

  /*     Intellon Specific MME format (p41)
   * 	MMECTRL   = 01
   * 	MMEHeader = 02
   * 	MMELength = 05
   *	MMEOUI    = 00 04 87
   *      Ox10 : Bit 7 = 0 (Request)
   *           Bits 6-0, MID = 0x10 (16 = Get Device Description)
   */
  qDebug("plc_send_mm : Get Device Description\n");
  sendto(netfd, mme, sizeof(mme), 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
  return 0;
}

int plc_broadcast_key()
{
  unsigned char mme[] = { 0x01, 0x14, 0x0A, 0x01, 0x00,
	                  0x9b, 0x1a, 0xa2, 0x0d, 0x83, 0xcb, 0x6e, 0xc8}; // always same data !! Tell me if you find what to put here ! TODO !

  /*     Intellon Specific MME format (p41)
   * 	MMECTRL   = 01
   * 	MEHDR = (MEV<<5) + MTYPE  = 0x14
   * 	MMELength = 0x0A
   *	MID       = 0x01 Encryption Key Tag Message ID
   *	EKS       = 0x00 Encryption Key Select
   */
  qDebug("plc_broadcast_key\n");
  sendto(netfd, mme, sizeof(mme), 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
  return 0;
}


int plc_get_device_version(const QString &mac)
{
  unsigned char mme[] = { 0x01, 0x12, 0x05, 0x00, 0x04, 0x87, 0x1};
  struct sockaddr_ll addr= { 0,0,0,0,0,0,{0,} };
  int i;

  /*     Intellon Specific MME format (p41)
   *      MMECTRL   = 01
   * 	MEHDR = (MEV<<5) + MTYPE  = 0x12
   *      MMELength = 05
   *      MMEOUI    = 00 04 87
   *      Ox10 : Bit 7 = 0 (Request)
   *           Bits 6-0, MID = 0x1 (1 = Get Device Version)
   */
  addr.sll_family  = AF_PACKET;
  addr.sll_protocol= htons(ETHERTYPE_INTELLON);
  addr.sll_ifindex = ifr.ifr_ifindex;

  std::wistringstream Stream(mac.toStdWString());

  for (i=0;i < 6; i++)
  {
    unsigned Byte;
    wchar_t Sep;
    if (i > 0)
    {
      Stream >> Sep;
    }
    Stream >> std::hex >> Byte;
    addr.sll_addr[i] = Byte;
  }

  qDebug("plc_send_req_version to ");
  for (i=0; i < 6; ++i)
  {
    qDebug("<%02x>",addr.sll_addr[i]);
  }
  qDebug("\n");
  qDebug("Get Device Version\n");
  sendto(netfd, mme, sizeof(mme), 0, (struct sockaddr *)&addr, sizeof(addr));
  return 0;
}


int plc_get_channel_capacities()
{
  unsigned char mme[] = { 0x01, 0x02, 0x06, 0x00, 0x04, 0x87, 0x18, 0x0};

  /*     Intellon Specific MME format (p44)
   * 	MMECTRL   = 01
   * 	MMEHeader = 02
   * 	MMELength = 05
   *	MMEOUI    = 00 04 87
   *      Ox18 : Bit 7 = 0 (Request)
   *           Bits 6-0, MID = 0x18 (24 = Get Channel Capacities)
   *      0x00 : (0 : Request Receive Capacities, 01 : transmit)
   *
   */
  qDebug("plc_send_mm : Get Channel capacities for receive\n");
  capabilities_flag = CAPABILITIES_RX;
  sendto(netfd, mme, sizeof(mme), 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
  sleep(1);
  mme[7] = 1; // Request transmit Capacities
  qDebug("plc_send_mm : Get Channel capacities for transmit\n");
  capabilities_flag = CAPABILITIES_TX;
  sendto(netfd, mme, sizeof(mme), 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
  return 0;
}

int plc_req_channel_estimation()
{
  unsigned char mme[] = { 0x01, 0x00, 0x01, 0x00};
  struct sockaddr_ll addr= { 0,0,0,0,0,0,{0,} };

  /*     Homeplug 1.0.1 format (p28)
   *     Request channel info format (p31)
   * 	MCTRL = NE     = 01
   * 	MEHDR = (MEV<<5) + MTYPE  = 00 : Request Channel Estimation
   * 	MELen = 01
   * 	MMENTRY = CEV       = 0 : INT5500
   */
  addr.sll_family  = AF_PACKET;
  addr.sll_protocol= htons(ETHERTYPE_INTELLON);
  addr.sll_ifindex = ifr.ifr_ifindex;
  addr.sll_addr[0] = 0x00;
  addr.sll_addr[1] = 0x0C;
  addr.sll_addr[2] = 0xb9;
  addr.sll_addr[3] = 0x03;
  addr.sll_addr[4] = 0x6f;
  addr.sll_addr[5] = 0xc1;

  qDebug("plc_req_channel_estimation\n");
  sendto(netfd, mme, sizeof(mme), 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
  //	sendto(netfd, mme, sizeof(mme), 0, (struct sockaddr *)&addr, sizeof(addr));
  return 0;
}

// read Firmware Version
int plc_req_version(const QString &mac)
{
  int i;
  struct sockaddr_ll addr= { 0,0,0,0,0,0,{0,} };
  unsigned char req[258] = {                                                                    0x01, 0x12,
    0x01, 0x20, 0x00, 0x00, 0xe5, 0x03, 0x00, 0x00,   0x68, 0xfb, 0x12, 0x00, 0x82, 0x93, 0x80, 0x7c,
    0xe5, 0x03, 0x00, 0x00, 0xb4, 0x88, 0x44, 0x00,   0xa8, 0xfb, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x16, 0x0e, 0x81, 0x7c, 0xf8, 0x00, 0x00, 0x00,   0xc8, 0x88, 0x44, 0x00, 0x3c, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   0xd0, 0xfb, 0x12, 0x00, 0x18, 0x1a, 0x80, 0x7c,
    0x74, 0xfb, 0x12, 0x00, 0x4c, 0x1f, 0x43, 0x00,   0xf8, 0xfc, 0x12, 0x00, 0xa8, 0x9a, 0x83, 0x7c,
    0x20, 0x0e, 0x81, 0x7c, 0xff, 0xff, 0xff, 0xff,   0x16, 0x0e, 0x81, 0x7c, 0xce, 0x4f, 0x00, 0x10,
    0x78, 0x00, 0x00, 0x00, 0xc8, 0x88, 0x44, 0x00,   0x3c, 0x00, 0x00, 0x00, 0xf4, 0xfb, 0x12, 0x00,
    0x00, 0xf0, 0xfd, 0x7f, 0x1c, 0xfc, 0x12, 0x00,   0x3c, 0x24, 0x80, 0x7c, 0xe0, 0xfb, 0x12, 0x00,
    0x0f, 0x24, 0x80, 0x7c, 0x9c, 0x92, 0x80, 0x7c,   0x77, 0x55, 0xb0, 0x01, 0x00, 0x00, 0x00, 0x00,
    0x14, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00,   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x10, 0x00, 0x00, 0x00, 0xf0, 0xd8, 0xff, 0xff,   0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
    0xf4, 0xfb, 0x12, 0x00, 0xd4, 0xfb, 0x12, 0x00,   0x60, 0x7d, 0x41, 0x00, 0xf8, 0xfc, 0x12, 0x00,
    0xa8, 0x9a, 0x83, 0x7c, 0x58, 0x24, 0x80, 0x7c,   0xff, 0xff, 0xff, 0xff, 0x0f, 0x24, 0x80, 0x7c,
    0x51, 0x24, 0x80, 0x7c, 0x01, 0x00, 0x00, 0x00,   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x76, 0xa4, 0x40, 0x00, 0x7d, 0xa4, 0x40, 0x00,   0x00, 0x00, 0x00, 0x00, 0x60, 0x7d, 0x41, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   0x03, 0xa0, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00
  };
  addr.sll_family  = AF_PACKET;
  addr.sll_protocol= htons(ETHERTYPE_INTELLON);
  addr.sll_ifindex = ifr.ifr_ifindex;

  std::wistringstream Stream(mac.toStdWString());

  for (i=0;i < 6; i++)
  {
    unsigned Byte;
    wchar_t Sep;
    if (i > 0)
    {
      Stream >> Sep;
    }
    Stream >> std::hex >> Byte;
    addr.sll_addr[i] = Byte;
  }

  qDebug("plc_send_req_version to ");
  for (i=0; i < 6; ++i)
  {
    qDebug("<%02x>",addr.sll_addr[i]);
  }
  qDebug("\n");
  sendto(netfd, req, sizeof(req), 0, (struct sockaddr *)&addr, sizeof(addr));
  return 0;
}

#endif
