/***************************************************************************
 *   Copyright (C) 2007 by Stephane List                                   *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef PLC_H
#define PLC_H

#include <QMainWindow>
#include <QCloseEvent>
#include <QLineEdit>
#include <QTabWidget>
#include <QTableWidget>
#include <QDialogButtonBox>
#include <QDialog>
#include <QGroupBox>
#include <QThread>
#include <QString>
#include <QSettings>
#include <QtDebug>
#include <QComboBox>

#if defined(Q_OS_WIN32)
#include <pcap.h>
#else
#include <arpa/inet.h>
#include <netpacket/packet.h> 
#include <net/ethernet.h>
#endif

#include <sstream>
#include <iomanip>


class QAction;
class QMenu;
class QTextEdit;

class MainTab;
class PrivacyTab;
class AboutTab;

enum {
	NEK_IDLE = 0,
	NEK_SENT_LOCAL,
	NEK_SENT_FIRST,
	NEK_SENT_SECOND
};
		
/* TODO : add something for a clean deinit... */
class plc:public QDialog
{
	Q_OBJECT

	public:
		plc();
		~plc();
		QString& getMyMac();
		void setMyMac(const QString& mac);
		bool autoscan;
		QSettings *settings;
		static void StaticHandlePacket(unsigned char* instance, const struct pcap_pkthdr *header, const u_char *packet);
		void HandlePacket(unsigned char *args, const struct pcap_pkthdr *header, const u_char *packet);
		void register_callback();

	public slots:
		void change_if(const QString &);

	signals:
		void Mac_detected(const QString & mac);
		void Date_detected(const QString & mac);
		void Rate_detected(const QString & mac, int rate, int dir);
		void My_Mac_detected(const QString & mac, const QString & firmware);
		void Nek_changed_confirm();

	private:
		void createTab();
		QTabWidget *tabWidget;
		QDialogButtonBox *buttonBox;
		QThread *scanThread;
		MainTab *MainTab_;
		PrivacyTab *PrivacyTab_;
		QGroupBox *IfGroup;
		int if_count;
		QString MyMac;
};

class MainTab : public QTabWidget
{
	Q_OBJECT

	public:
		MainTab(plc *parent);
		plc* Parent;
		QTableWidget *ConnectedTable;
		QTableWidget *ScannedTable;

	public slots:
		void add_mac(const QString & mac);
		void add_my_mac(const QString & mac, const QString & firmware);
		void add_rate(const QString & mac, int rate, int dir);
		void add_date(const QString & mac);
		void remove_old();
};


class PrivacyTab : public QTabWidget
{
	Q_OBJECT
	public:
		PrivacyTab(plc *parent);
		plc* Parent;

	private:
		QLineEdit *nekLineEdit;
		char Nek_state;

	public slots:
		void setDefaultNek();
		void setLocalDevice();
		void setAllDevices();
		void display_nek_confirm();
};

class AboutTab : public QTabWidget
{
	Q_OBJECT

	public:
		AboutTab(plc *parent);
		plc* Parent;

	public slots:
		void AutoScan_toggled(bool val);

};

class ScanThread : public QThread
{
	Q_OBJECT

	public:
		ScanThread(plc *parent = 0);
		void run();

	signals:
		void remove_old();

};


class pcapThread : public QThread
{
	Q_OBJECT

	public:
		pcapThread(plc *parent = 0);
		void run();
};

#ifndef ETHER_ADDR_LEN
#define ETHER_ADDR_LEN 6
#endif

/* Ethernet header */
struct sniff_ethernet {
	u_char ether_dhost[ETHER_ADDR_LEN]; /* Destination host address */
	u_char ether_shost[ETHER_ADDR_LEN]; /* Source host address */
	u_short ether_type; /* IP? ARP? RARP? etc */
};

#endif
